/*	Copyright (C) 1992 Free Software Foundation, Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this software; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */


#include "region.h"


xx_Region
xx_region ()
{
  return xx_rtree ();
}

void
xx_free_region (r)
     xx_Region r;
{
  xx_free_rtree (r);
}

int
xx_region_empty (r)
     xx_Region r;
{
  return !r->root->children[0];
}


static int
rect_copy (leaf, rect, output)
     xx_RtreeLeaf leaf;
     xx_IntRectangle rect;
     void *output;
{
  xx_RtreeLeaf *out = output;
  *out = xx_rtreeleaf ((void *) 23);
  return 1;
}

xx_Region
xx_region_copy (r)
     xx_Region r;
{
  return xx_rtree_copy (r, rect_copy);
}


struct rect_change
{
  xx_RtreeLeaf leaf;
  struct xx_sIntRectangle rect;
  int delete;
  struct rect_change *next;
};

struct region_edit_frame
{
  struct rect_change *changes;
  struct xx_sIntRectangle arg;
};

static struct rect_change *
rect_change (leaf, rect, delete, list)
     xx_RtreeLeaf leaf;
     xx_IntRectangle rect;
     int delete;
     struct rect_change *list;
{
  struct rect_change *r = (struct rect_change *) ck_malloc (sizeof (*r));
  r->leaf = leaf;
  r->rect = *rect;
  r->next = list;
  r->delete = delete;
  return r;
}

static void
build_changes (frame, leaf, leafrect, rectc, rectv)
     struct region_edit_frame *frame;
     xx_RtreeLeaf *leaf;
     xx_IntRectangle leafrect;
     int rectc;
     xx_IntRectangle rectv;
{
  if (!rectc)
    frame->changes = rect_change (leaf, leafrect, 1, frame->changes);
  else
    {
      frame->changes = rect_change (leaf, rectv, 0, frame->changes);
      ++rectv;
      --rectc;

      while (rectc--)
	{
	  xx_RtreeLeaf newleaf = xx_rtreeleaf ((void *) 0xface);
	  frame->changes = rect_change (newleaf, rectv, 0, frame->changes);
	  ++rectv;
	}
    }
}

static void
free_rect_change (list)
     struct rect_change *list;
{
  if (list->next)
    free_rect_change (list->next);
  ck_free (list);
}

static void
do_changes (r, list)
     xx_Region r;
     struct rect_change *list;
{
  while (list)
    {
      struct rect_change *temp = list;
      list = list->next;
      temp->next = 0;

      if (temp->delete)
	{
	  xx_rtree_delete (r, temp->leaf);
	  xx_free_rtreeleaf (temp->leaf);
	}
      else
	{
	  if (temp->leaf->parent)
	    xx_rtree_delete (r, temp->leaf);
	  xx_rtree_add (r, &temp->rect, temp->leaf);
	}
      free_rect_change (temp);
    }
}

static int
subtract (leaf, rect, voidf)
     xx_RtreeLeaf leaf;
     xx_IntRectangle rect;
     void *voidf;
{
  struct region_edit_frame *frame = (struct region_edit_frame *) voidf;
  struct xx_sIntRectangle region_rect;
  struct xx_sIntRectangle arg_rect;
  struct xx_sIntRectangle output_rects[4];
  int outp = 0;
  region_rect = *rect;

  arg_rect = frame->arg;
  xx_IRclip (&arg_rect, &region_rect);

  if (xx_IRhits_point (&region_rect,
		       xx_IRllx (&arg_rect),
		       xx_IRlly (&arg_rect)))
    {
      /*

                       arg_rect
	 -----------------
	|        |////////|
	|        |////////|
        |        |---------
	|   I    |  II    |
	| 	 |        |
         -----------------  region_rect
	*/

      /* I  */
      output_rects[outp] = region_rect;
      output_rects[outp].w = xx_IRllx (&arg_rect) - xx_IRllx (&region_rect);

      if (output_rects[outp].w)
	++outp;

      /* II */
      output_rects[outp] = region_rect;
      output_rects[outp].x = xx_IRllx (&arg_rect);
      output_rects[outp].w = xx_IRurx (&region_rect) - xx_IRllx (&arg_rect) + 1;
      output_rects[outp].h = xx_IRlly (&arg_rect) - xx_IRlly (&region_rect);

      if (output_rects[outp].h)
	++outp;

      region_rect.w = xx_IRurx (&region_rect) - xx_IRllx (&arg_rect) + 1;
      region_rect.h = xx_IRury (&region_rect) - xx_IRlly (&arg_rect) + 1;
      region_rect.x = xx_IRllx (&arg_rect);
      region_rect.y = xx_IRlly (&arg_rect);
    }

  if (xx_IRhits_point (&region_rect,
		       xx_IRurx (&arg_rect),
		       xx_IRury (&arg_rect)))
    {
      /*
	          --------------
                 |           I  | region_rect
                 |--------------|
                 |////////|     |
                 |////////|  II |
                  --------------
             arg_rect


	*/

      /* I */
      output_rects[outp] = region_rect;
      output_rects[outp].h = xx_IRury (&region_rect) - xx_IRury (&arg_rect);
      output_rects[outp].y = xx_IRury (&arg_rect) + 1;
      if (output_rects[outp].h)
	++outp;

      /* II */
      output_rects[outp] = region_rect;
      output_rects[outp].h = xx_IRury (&arg_rect) - xx_IRlly (&region_rect);
      output_rects[outp].w = xx_IRurx (&region_rect) - xx_IRurx (&arg_rect);
      output_rects[outp].x = xx_IRurx (&arg_rect) + 1;

      if (output_rects[outp].w)
	++outp;
    }

  build_changes (frame, leaf, &region_rect, outp, output_rects);
  return 1;
}

void
xx_region_subtract_rectangle (r, rect)
     xx_Region r;
     xx_IntRectangle rect;
{
  struct region_edit_frame f;

  f.changes = 0;
  f.arg = *rect;

  xx_rtree_query (r, rect, subtract, &f);
  do_changes (r, f.changes);
}

static int
intersect (leaf, rect, voidf)
     xx_RtreeLeaf leaf;
     xx_IntRectangle rect;
     void *voidf;
{
  struct region_edit_frame *frame = (struct region_edit_frame *) voidf;
  struct xx_sIntRectangle output_rect;
  struct xx_sIntRectangle arg_rect;

  output_rect = *rect;

  arg_rect = frame->arg;
  xx_IRclip (&arg_rect, &output_rect);

  if (xx_IRhits_point (&output_rect,
		       xx_IRllx (&arg_rect),
		       xx_IRlly (&arg_rect)))
    {
      /*

	                  arg_rect
	 -----------------
	|        |////////|
	|        |////////|
        |        |---------
	|                 |
	| 	          |
         -----------------  output_rect
	*/

      output_rect.w = xx_IRurx (rect) - xx_IRllx (&arg_rect) + 1;
      output_rect.h = xx_IRury (rect) - xx_IRlly (&arg_rect) + 1;
      output_rect.x = xx_IRllx (&arg_rect);
      output_rect.y = xx_IRlly (&arg_rect);
    }

  if (xx_IRhits_point (&output_rect,
		       xx_IRurx (&arg_rect),
		       xx_IRury (&arg_rect)))
    {
      /*
	          --------------
                 |              | output_rect
                 |--------      |
                 |XXXXXXXX|     |
                 |XXXXXXXX|     |
                  --------------
             arg_rect
	*/

      output_rect.h = xx_IRury (&arg_rect) - xx_IRlly (&output_rect) + 1;
      output_rect.w = xx_IRurx (&arg_rect) - xx_IRllx (&output_rect) + 1;
    }

  frame->changes = rect_change (leaf, &output_rect,
				!(output_rect.w || output_rect.h),
				frame->changes);
  return 1;
}

void
xx_region_intersect_rectangle (r, rect)
     xx_Region r;
     xx_IntRectangle rect;
{
  struct rect_change *changes = 0;

  xx_rtree_query (r, rect, intersect, &changes);
  do_changes (r, changes);
}


struct expand_frame
{
  xx_RtreeLeaf leaf;
  int area;
  struct xx_sIntRectangle rectangle;
  struct xx_sIntRectangle new_rect;
};

#define MIN(A,B) (((A) < (B)) ? (A) : (B))
#define MAX(A,B) (((A) > (B)) ? (A) : (B))

static int
expand (leaf, rect, voidexpand)
     xx_RtreeLeaf leaf;
     xx_IntRectangle rect;
     void *voidexpand;
{
  struct expand_frame *frame = (struct expand_frame *) voidexpand;
  struct xx_sIntRectangle frob_rect;

  frob_rect.x = MIN (rect->x, frame->new_rect.x);
  frob_rect.y = MIN (rect->y, frame->new_rect.y);
  frob_rect.w =
    MAX (xx_IRurx (rect), xx_IRurx (&frame->new_rect))
    - frob_rect.x + 1;
  frob_rect.h =
    MAX (xx_IRury (rect), xx_IRury (&frame->new_rect))
    - frob_rect.y + 1;


  {
    int frob_area = xx_IRarea (&frob_rect);
    if ((frob_area > frame->area) &&
	(frob_area ==
	 xx_IRarea (&frame->new_rect) + xx_IRarea (&frame->rectangle)))
      {
	frame->rectangle = *rect;
	frame->area = frob_area;
      }
  }
  return 1;
}

void
xx_region_adjoin_rectangle (r, rect)
     xx_Region r;
     xx_IntRectangle rect;
{
  struct xx_sIntRectangle border_rect;
  struct expand_frame expand_frame;

  expand_frame.leaf = 0;
  expand_frame.area = 0;

  xx_region_subtract_rectangle (r, rect);
  border_rect.x = rect->x - 1;
  border_rect.y = rect->y - 1;
  border_rect.w = rect->w + 2;
  border_rect.h = rect->h + 2;
  xx_rtree_query (r, &border_rect, expand, &expand_frame);

  if (expand_frame.area)
    {
      xx_rtree_delete (r, expand_frame.leaf);
      xx_free_rtreeleaf (expand_frame.leaf);
      xx_region_adjoin_rectangle (r, &expand_frame.rectangle);
    }
  else
    {
      xx_RtreeLeaf leaf = xx_rtreeleaf ((void *) 123456);
      xx_rtree_add (r, rect, leaf);
    }
}

static int
unioner (rect, vdest)
     xx_IntRectangle rect;
     void *vdest;
{
  xx_Region dest = (xx_Region) vdest;

  xx_region_adjoin_rectangle (dest, rect);
}

void
xx_region_union (dest, src)
     xx_Region dest;
     xx_Region src;
{
  xx_region_iterate (src, unioner, dest);
}

struct adj_loosely_frame
{
  struct xx_sIntRectangle total_rect;
  struct xx_sIntRectangle query_rect;
};

#define ISMOST(A,B)  !(10 * ((B) - (A)) > (B))

static int
loose_rect_combiner (leaf, regrect, voidframe)
     xx_RtreeLeaf leaf;
     xx_IntRectangle regrect;
     void *voidframe;
{
  struct adj_loosely_frame *frame = (struct adj_loosely_frame *) voidframe;
  struct xx_sIntRectangle new_regrect;
  new_regrect = *regrect;

  xx_IRbound (&new_regrect, regrect);
  if (ISMOST (xx_IRarea (regrect), xx_IRarea (&new_regrect))
      && ISMOST (xx_IRarea (&frame->query_rect), xx_IRarea (&new_regrect)))
    xx_IRbound (&frame->total_rect, regrect);
  return 1;
}


void
xx_region_adjoin_rectangle_loosely (r, rect)
     xx_Region r;
     xx_IntRectangle rect;
{
  struct adj_loosely_frame frame;
  frame.query_rect = *rect;
  frame.total_rect = *rect;
  xx_rtree_query (r, rect, loose_rect_combiner, (void *) &frame);
  xx_region_adjoin_rectangle (r, &frame.total_rect);
}


struct iterate_frame
{
  xx_RegionIterator fn;
  void *data;
};

static int
region_iter (leaf, rect, data)
     xx_RtreeLeaf leaf;
     xx_IntRectangle rect;
     void *data;
{
  struct iterate_frame *frame = (struct iterate_frame *) data;
  return (frame->fn) (rect, frame->data);
}

void
xx_region_iterate (r, iter, data)
     xx_Region r;
     xx_RegionIterator iter;
     void *data;
{
  struct iterate_frame f;
  f.fn = iter;
  f.data = data;
  xx_rtree_query (r, &r->domain, region_iter, &f);
}

void
xx_region_clear (r)
     xx_Region r;
{
  struct xx_sIntRectangle range;
  range = r->domain;
  xx_region_subtract_rectangle (r, &range);
}



int
xx_region_encloses (r1, r2)
     xx_Region r1;
     xx_Region r2;
{
  if (xx_region_empty (r2))
    return 1;

  if (xx_region_empty (r1))
    return 0;

  if (xx_IRencloses (&r1->domain, &r2->domain))
    {
      xx_Region r1_copy = xx_region_copy (r1);
      int result;
      xx_region_subtract_region (r1_copy, r2);
      result = xx_region_empty (r1_copy);
      xx_free_region (r1_copy);
      return result;
    }

  return 0;
}

int
xx_region_equiv (r1, r2)
     xx_Region r1;
     xx_Region r2;
{
  return (r1 == r2
	  || (xx_IRequiv (&r1->domain, &r2->domain) &&
	      xx_IRencloses (&r1->domain, &r2->domain) &&
	      xx_IRencloses (&r2->domain, &r1->domain)));
}

static int
intersects_rect_loopfn (leaf, rect, voidint)
     xx_RtreeLeaf leaf;
     xx_IntRectangle rect;
     void *voidint;
{
  int *i = (int *) voidint;

  *i = 1;
  return 0;
}

int
xx_region_intersects_rectangle (r, rect)
     xx_Region r;
     xx_IntRectangle rect;
{
  int hit = 0;

  xx_rtree_query (r, rect, intersects_rect_loopfn, &hit);
  return hit;
}

struct region_frame
{
  xx_Region key;
  int result;
};

static int
intersects_region_loopfn (rect, voidframe)
     xx_IntRectangle rect;
     void *voidframe;
{
  struct region_frame *frame =
  (struct region_frame *) voidframe;

  if (xx_region_intersects_rectangle (frame->key, rect))
    {
      frame->result = 1;
      return 0;
    }
  else
    return 1;
}

int
xx_region_intersects (r1, r2)
     xx_Region r1;
     xx_Region r2;
{
  struct region_frame frame;
  frame.key = r2;
  frame.result = 0;
  xx_region_iterate (r1, intersects_region_loopfn, &frame);
  return frame.result;
}

static int
subtractrg (rect, voidfr)
     xx_IntRectangle rect;
     void *voidfr;
{
  struct region_frame *frame = (struct region_frame *) voidfr;
  xx_region_subtract_rectangle (frame->key, rect);
  return 1;
}

void
xx_region_subtract_region (r1, r2)
     xx_Region r1;
     xx_Region r2;
{
  struct region_frame frame;
  frame.key = r1;
  xx_region_iterate (r2, subtractrg, &frame);
}



static int
intersectrg (rect, voidfr)
     xx_IntRectangle rect;
     void *voidfr;
{
  struct region_frame *frame = (struct region_frame *) voidfr;
  xx_region_intersect_rectangle (frame->key, rect);
  return 1;
}

void
xx_region_intersect_region (r1, r2)
     xx_Region r1;
     xx_Region r2;
{
  struct region_frame frame;
  frame.key = r1;
  xx_region_iterate (r2, intersectrg, &frame);
}


static int
addup_area (rect, vint)
     xx_IntRectangle rect;
     void *vint;
{
  int *i = (int *) vint;

  *i += xx_IRarea (rect);
  return 1;
}

unsigned int
xx_region_area (reg)
     xx_Region reg;
{
  int total = 0;
  xx_region_iterate (reg, addup_area, &total);
  return total;
}

void
xx_region_complement (reg, rect)
     xx_Region reg;
     xx_IntRectangle rect;
{
  static xx_Region r2 = 0;
  if (!r2)
    r2 = xx_region ();
  xx_region_adjoin_rectangle (r2, rect);
  xx_region_subtract_region (r2, reg);
  xx_region_subtract_rectangle (reg, rect);
  xx_region_union (reg, r2);
  xx_region_clear (r2);
}
